{-# OPTIONS --without-K --type-in-type #-}

open import Agda.Primitive renaming (_⊔_ to lmax) public
open import Agda.Builtin.Bool public
open import Agda.Builtin.Sigma public
open import Agda.Builtin.Unit public
open import Agda.Builtin.Equality public
open import Agda.Builtin.Nat renaming (Nat to ℕ) public
open import Agda.Builtin.FromNat public
open import Agda.Builtin.FromNeg public

data TLevel : Set where
  ⟨-2⟩ : TLevel
  suc : (t : TLevel) → TLevel

postulate
  magic : ∀ {i} {X : Set i} → X

Type : Set₁
Type = Set -- a better name for Set

data _⊔_ (A : Type) (B : Type) : Type where
  inl : A → A ⊔ B
  inr : B → A ⊔ B
infixr 4 _⊔_

data ⊥ : Type where

abort : {A : Type} → ⊥ → A
abort ()

¬_ : Type → Type
¬ A = A → ⊥

{- START of Agda hack to write numerals directly -}
instance
  Numℕ : Number ℕ
  Number.Constraint Numℕ _ = ⊤
  Number.fromNat    Numℕ n = n
  NumTLevel : Number TLevel
  Number.Constraint NumTLevel _ = ⊤
  Number.fromNat    NumTLevel k = natToTLevel k
    where
      natToTLevel : ∀ ℕ → {{_ : ⊤}} → TLevel
      natToTLevel 0 = suc (suc ⟨-2⟩)
      natToTLevel (suc n) = suc (natToTLevel n)
  NegTLevel : Negative TLevel
  Negative.Constraint NegTLevel 0 = ⊤
  Negative.Constraint NegTLevel 1 = ⊤
  Negative.Constraint NegTLevel 2 = ⊤
  Negative.Constraint NegTLevel _ = ⊥
  Negative.fromNeg    NegTLevel 0 = suc (suc ⟨-2⟩)
  Negative.fromNeg    NegTLevel 1 = suc ⟨-2⟩
  Negative.fromNeg    NegTLevel 2 = ⟨-2⟩
  Negative.fromNeg    NegTLevel (suc (suc (suc _))) ⦃()⦄
{- END of the Agda hack to write numerals -}

ap : ∀ {A B : Type} (f : A → B) {x y : A} → x ≡ y → f x ≡ f y
ap f refl = refl

{- groupoidal aspects -}

_∙_ : ∀ {A : Type} {x y z : A} → x ≡ y → y ≡ z → x ≡ z
refl ∙ q = q
infixr 80 _∙_

!_ : ∀ {A : Type} {x y : A} → x ≡ y → y ≡ x
! refl = refl
infixr 100 !_

∙-unit-r : ∀ {A : Type} {x y : A} (q : x ≡ y) → q ∙ refl ≡ q
∙-unit-r refl = refl

∙-unit-l : ∀ {A : Type} {x y : A} (q : x ≡ y) → refl ∙ q ≡ q
∙-unit-l q = refl

!-inv-l : ∀ {A : Type} {x y : A} (p : x ≡ y) → (! p) ∙ p ≡ refl
!-inv-l refl = refl

!-inv-r : ∀ {A : Type} {x y : A} (p : x ≡ y) → p ∙ (! p) ≡ refl
!-inv-r refl = refl

!-! : ∀ {A : Type} {x y : A} (p : x ≡ y) → ! (! p) ≡ p
!-! refl = refl

∙-assoc : ∀ {A : Type} {x y z w : A} (p : x ≡ y) (q : y ≡ z) (r : z ≡ w)
  → (p ∙ q) ∙ r ≡ p ∙ (q ∙ r)
∙-assoc refl q r = refl

{- functorial aspects -}

ap-∙ : ∀ {A B : Type} (f : A → B) {x y z : A} (p : x ≡ y) (q : y ≡ z)
  → ap f (p ∙ q) ≡ ap f p ∙ ap f q
ap-∙ f refl q = refl

ap-! : ∀ {A B : Type} (f : A → B) {x y : A} (p : x ≡ y) → ap f (! p) ≡ ! (ap f p)
ap-! f refl = refl

{- we like truncation levels more today -}

is-contr : Type → Type
is-contr A = Σ A (λ x → (y : A) → x ≡ y)

has-level : TLevel → Type → Type
has-level ⟨-2⟩ A = is-contr A
has-level (suc n) A = (x y : A) → has-level n (x ≡ y)

is-prop : Type → Type
is-prop = has-level -1

is-set : Type → Type
is-set = has-level 0

record is-equiv {A B : Type} (f : A → B) : Type where
  field
    g : B → A
    f-g : ∀ (b : B) → f (g b) ≡ b
    g-f : ∀ (a : A) → g (f a) ≡ a
    adj : ∀ (a : A) → ap f (g-f a) ≡ f-g (f a)

infix 30 _≃_
_≃_ : Type → Type → Type
A ≃ B = Σ (A → B) is-equiv

{- quasi-inverses -> equivalences -}
is-eq : ∀ {A B : Type} (f : A → B) (g : B → A) (f-g : (y : B) → f (g y) ≡ y) (g-f : (x : A) → g (f x) ≡ x) → is-equiv f
is-eq {A} {B} f g f-g g-f =
  record {g = g; f-g = f-g'; g-f = g-f; adj = adj} where
    abstract
      {- adjusting f-g -}
      f-g' : (y : B) → f (g y) ≡ y
      f-g' b = ! (ap (λ y → f (g y)) (f-g b)) ∙ ap f (g-f (g b)) ∙ f-g b

      {- you can prove this with adjusted f-g, but a full proof here would
         make this file unnecessarily long. -}
      postulate
        adj : (a : A) → ap f (g-f a) ≡ f-g' (f a)

equiv : ∀ {A B : Type} (f : A → B) (g : B → A) (f-g : (y : B) → f (g y) ≡ y) (g-f : (x : A) → g (f x) ≡ x) → A ≃ B
equiv f g f-g g-f = f , is-eq f g f-g g-f

–> : ∀ {A B : Type} → A ≃ B → (A → B)
–> e a = fst e a

<– : ∀ {A B : Type} → A ≃ B → (B → A)
<– e b = is-equiv.g (snd e) b

<–-inv-l :  ∀ {A B : Type} → (e : A ≃ B) (a : A)
  → (<– e (–> e a) ≡ a)
<–-inv-l e a = is-equiv.g-f (snd e) a

<–-inv-r :  ∀ {A B : Type} → (e : A ≃ B) (b : B)
  → (–> e (<– e b) ≡ b)
<–-inv-r e b = is-equiv.f-g (snd e) b

_×_ : Type → Type → Type
A × B = Σ A (λ _ → B)

{- START of the tasks -}

task1 : is-contr ⊤
task1 = magic

task2 : is-prop ⊥
task2 = magic

task3-1 : ∀ (x y : Bool) → Type
task3-1 = magic
{- See the type of task3-2. -}

task3-2 : ∀ (x y : Bool) → (x ≡ y) ≃ (task3-1 x y)
task3-2 = magic
{- You are forbidden to use the absurd pattern "()" directly in task3-2.
   You can only use task3-1, lemmas about equivalences, and "abort".
   See ≡⊔-econv and ≡ℕ-econv in the Agda self-practice scheduled on 2/27. -}

task3-3 : ¬ (true ≡ false)
task3-3 = magic
{- You are forbidden to use the absurd pattern "()" directly in task3-3.
   You can only use task3-1, task3-2, lemmas about equivalences, and "abort".
   See ¬0≡suc in the Agda self-practice scheduled on 2/27. -}

{- Two *potentially* useful lemmas for Task 4. -}
postulate
  has-level-preserves-equiv : ∀ n {A B : Type} → A ≃ B → has-level n A → has-level n B
  has-level-preserves'-equiv : ∀ n {A B : Type} → A ≃ B → has-level n B → has-level n A

≡× : ∀ {A B : Type} (x y : A × B) → Type
≡× x y = (fst x ≡ fst y) × (snd x ≡ snd y)

postulate
  ≡×-econv : ∀ {A B : Type} (x y : A × B) → (x ≡ y) ≃ (≡× x y)

task4 : ∀ (n : TLevel) (A B : Type)
  → has-level n A → has-level n B → has-level n (A × B)
task4 = magic
{- Hints: Did you finish raise-level in earlier Agda practices yet? If not, please do that first. -}

happly : ∀ {A : Type} {B : A → Type} {f g : (x : A) → B x} (p : f ≡ g) → (x : A) → f x ≡ g x
happly p x = ap (λ f → f x) p

{- STOP if you have not done the Agda self-practice scheduled on 2/27 (but was delayed). -}

postulate
  funext : ∀ {A : Type} {B : A → Type} {f g : (x : A) → B x} → is-equiv (happly {f = f} {g = g})

≡Π : ∀ {A : Type} {B : A → Type} (f g : (x : A) → B x) → Type
≡Π {A = A} f g = (x : A) → f x ≡ g x

≡Π-econv : ∀ {A : Type} {B : A → Type} (f g : (x : A) → B x) → (f ≡ g) ≃ ≡Π f g
≡Π-econv f g = happly , funext

{- this is the most used part of the function extensionality -}
λ≡ : ∀ {A : Type} {B : A → Type} {f g : (x : A) → B x} → ≡Π f g → f ≡ g
λ≡ {f = f} {g = g} h = <– (≡Π-econv f g) h

task5 : ∀ (n : TLevel) {A : Type} (B : A → Type)
  {- A is made implicit because it can be inferred from B in 99% of the cases. -}
  → ((x : A) → has-level n (B x)) → has-level n ((x : A) → B x)
task5 = magic

-- Remember to remove all "magic"!
